@@ -1,10 +1,10 @@
 /*  GNU SED, a batch stream editor.
-    Copyright (C) 1989,90,91,92,93,94,95,98,99,2002,2003,2004,2005,2006
+    Copyright (C) 1989,90,91,92,93,94,95,98,99,2002,2003,2004,2005,2006,2008,2010
     Free Software Foundation, Inc.
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
-    the Free Software Foundation; either version 2, or (at your option)
+    the Free Software Foundation; either version 3, or (at your option)
     any later version.
 
     This program is distributed in the hope that it will be useful,
@@ -19,30 +19,11 @@
 /* compile.c: translate sed source into internal form */
 
 #include "sed.h"
-#include "strverscmp.h"
 #include <stdio.h>
 #include <ctype.h>
-
-#ifdef HAVE_STRINGS_H
-# include <strings.h>
-# ifdef HAVE_MEMORY_H
-#  include <memory.h>
-# endif
-#else
-# include <string.h>
-#endif /* HAVE_STRINGS_H */
-
-#ifdef HAVE_STDLIB_H
-# include <stdlib.h>
-#endif
-#ifndef EXIT_FAILURE
-# define EXIT_FAILURE 1
-#endif
-
-#ifdef HAVE_SYS_TYPES_H
-# include <sys/types.h>
-#endif
-
+#include <string.h>
+#include <stdlib.h>
+#include <sys/types.h>
 #include <obstack.h>
 
 
@@ -156,7 +137,8 @@ static const char errors[] =
   "delimiter character is not a single-byte character\0"
   "expected newer version of sed\0"
   "invalid usage of line address 0\0"
-  "unknown command: `%c'";
+  "unknown command: `%c'\0"
+  "incomplete command";
 
 #define BAD_BANG (errors)
 #define BAD_COMMA (BAD_BANG + sizeof(N_("multiple `!'s")))
@@ -183,7 +165,8 @@ static const char errors[] =
 #define ANCIENT_VERSION (BAD_DELIM + sizeof(N_("delimiter character is not a single-byte character")))
 #define INVALID_LINE_0 (ANCIENT_VERSION + sizeof(N_("expected newer version of sed")))
 #define UNKNOWN_CMD (INVALID_LINE_0 + sizeof(N_("invalid usage of line address 0")))
-#define END_ERRORS (UNKNOWN_CMD + sizeof(N_("unknown command: `%c'")))
+#define INCOMPLETE_CMD (UNKNOWN_CMD + sizeof(N_("unknown command: `%c'")))
+#define END_ERRORS (INCOMPLETE_CMD + sizeof(N_("incomplete command")))
 
 static struct output *file_read = NULL;
 static struct output *file_write = NULL;
@@ -207,12 +190,12 @@ bad_prog(why)
 {
   if (cur_input.name)
     fprintf(stderr, _("%s: file %s line %lu: %s\n"),
-	    myname, cur_input.name, CAST(unsigned long)cur_input.line, why);
+	    myname, cur_input.name, (unsigned long)cur_input.line, why);
   else
     fprintf(stderr, _("%s: -e expression #%lu, char %lu: %s\n"),
 	    myname,
-	    CAST(unsigned long)cur_input.string_expr_count,
-	    CAST(unsigned long)(prog.cur-prog.base),
+	    (unsigned long)cur_input.string_expr_count,
+	    (unsigned long)(prog.cur-prog.base),
 	    why);
   exit(EXIT_FAILURE);
 }
@@ -221,7 +204,7 @@ bad_prog(why)
 /* Read the next character from the program.  Return EOF if there isn't
    anything to read.  Keep cur_input.line up to date, so error messages
    can be meaningful. */
-static int inchar P_((void));
+static int inchar (void);
 static int
 inchar()
 {
@@ -243,7 +226,7 @@ inchar()
 }
 
 /* unget `ch' so the next call to inchar will return it.   */
-static void savchar P_((int ch));
+static void savchar (int ch);
 static void
 savchar(ch)
   int ch;
@@ -256,14 +239,14 @@ savchar(ch)
     {
       if (prog.cur <= prog.base || *--prog.cur != ch)
 	panic("Called savchar() with unexpected pushback (%x)",
-	      CAST(unsigned char)ch);
+	      (unsigned char)ch);
     }
   else
     ungetc(ch, prog.file);
 }
 
 /* Read the next non-blank character from the program.  */
-static int in_nonblank P_((void));
+static int in_nonblank (void);
 static int
 in_nonblank()
 {
@@ -275,7 +258,7 @@ in_nonblank()
 }
 
 /* Read an integer value from the program.  */
-static countT in_integer P_((int ch));
+static countT in_integer (int ch);
 static countT
 in_integer(ch)
   int ch;
@@ -291,7 +274,7 @@ in_integer(ch)
   return num;
 }
 
-static int add_then_next P_((struct buffer *b, int ch));
+static int add_then_next (struct buffer *b, int ch);
 static int
 add_then_next(b, ch)
   struct buffer *b;
@@ -301,20 +284,19 @@ add_then_next(b, ch)
   return inchar();
 }
 
-static char * convert_number P_((char *, char *, const char *, int, int, int));
+static char * convert_number (char *, char *, const char *, int);
 static char *
-convert_number(result, buf, bufend, base, maxdigits, default_char)
+convert_number(result, buf, bufend, base)
   char *result;
   char *buf;
   const char *bufend;
   int base;
-  int maxdigits;
-  int default_char;
 {
   int n = 0;
+  int max = 1;
   char *p;
 
-  for (p=buf; p < bufend && maxdigits-- > 0; ++p)
+  for (p=buf+1; p < bufend && max <= 255; ++p, max *= base)
     {
       int d = -1;
       switch (*p)
@@ -340,8 +322,8 @@ convert_number(result, buf, bufend, base, maxdigits, default_char)
 	break;
       n = n * base + d;
     }
-  if (p == buf)
-    *result = default_char;
+  if (p == buf+1)
+    *result = *buf;
   else
     *result = n;
   return p;
@@ -349,7 +331,7 @@ convert_number(result, buf, bufend, base, maxdigits, default_char)
 
 
 /* Read in a filename for a `r', `w', or `s///w' command. */
-static struct buffer *read_filename P_((void));
+static struct buffer *read_filename (void);
 static struct buffer *
 read_filename()
 {
@@ -374,17 +356,16 @@ read_filename()
   return b;
 }
 
-static struct output *get_openfile P_((struct output **file_ptrs, char *mode, bool fail));
+static struct output *get_openfile (struct output **file_ptrs, const char *mode, int fail);
 static struct output *
 get_openfile(file_ptrs, mode, fail)
      struct output **file_ptrs;
-     char *mode;
-     bool fail;
+     const char *mode;
+     int fail;
 {
   struct buffer *b;
   char *file_name;
   struct output *p;
-  int is_stderr;
 
   b = read_filename();
   file_name = get_buffer(b);
@@ -423,7 +404,7 @@ get_openfile(file_ptrs, mode, fail)
 }
 
 
-static struct sed_cmd *next_cmd_entry P_((struct vector **vectorp));
+static struct sed_cmd *next_cmd_entry (struct vector **vectorp);
 static struct sed_cmd *
 next_cmd_entry(vectorp)
   struct vector **vectorp;
@@ -449,7 +430,7 @@ next_cmd_entry(vectorp)
   return cmd;
 }
 
-static int snarf_char_class P_((struct buffer *b, mbstate_t *cur_stat));
+static int snarf_char_class (struct buffer *b, mbstate_t *cur_stat);
 static int
 snarf_char_class(b, cur_stat)
   struct buffer *b;
@@ -491,10 +472,10 @@ snarf_char_class(b, cur_stat)
 	  if (state == 1)
 	    {
 	      delim = ch;
-	      state++;
+	      state = 2;
 	    }
-	  else if (ch == delim && state == 2)
-	    state++;
+	  else if (state == 2 && ch == delim)
+	    state = 3;
 	  else
 	    break;
 
@@ -504,7 +485,8 @@ snarf_char_class(b, cur_stat)
 	  if (pending_mb)
 	    continue;
 
-	  state++;
+	  if (state == 0)
+	    state = 1;
 	  continue;
 
 	case CLOSE_BRACKET:
@@ -529,20 +511,19 @@ snarf_char_class(b, cur_stat)
     }
 }
 
-static struct buffer *match_slash P_((int slash, bool regex));
+static struct buffer *match_slash (int slash, int regex);
 static struct buffer *
 match_slash(slash, regex)
   int slash;
-  bool regex;
+  int regex;
 {
   struct buffer *b;
   int ch;
-  bool pending_mb = false;
   mbstate_t cur_stat;
 
   memset (&cur_stat, 0, sizeof (mbstate_t));
 
-  if (BRLEN (slash, &cur_stat) == -2)
+  /* We allow only 1 byte characters for a slash.  */
   if (BRLEN (slash, &cur_stat) == -2)
     bad_prog (BAD_DELIM);
 
@@ -551,10 +532,8 @@ match_slash(slash, regex)
   b = init_buffer();
   while ((ch = inchar()) != EOF && ch != '\n')
     {
-      pending_mb = BRLEN (ch, &cur_stat) != 1;
-      pending_mb = BRLEN (ch, &cur_stat) != 1;
-
-      if (!pending_mb)
+      bool pending_mb = !MBSINIT (&cur_stat);
+      if (BRLEN (ch, &cur_stat) == 1 && !pending_mb)
 	{
 	  if (ch == slash)
 	    return b;
@@ -567,7 +546,7 @@ match_slash(slash, regex)
 	      else if (ch == 'n' && regex)
 	        ch = '\n';
 #endif
-	      else if (ch != '\n' && ch != slash)
+	      else if (ch != '\n' && (ch != slash || (!regex && ch == '&')))
 	        add1_buffer(b, '\\');
 	    }
           else if (ch == OPEN_BRACKET && regex)
@@ -588,7 +567,7 @@ match_slash(slash, regex)
   return NULL;
 }
 
-static int mark_subst_opts P_((struct subst *cmd));
+static int mark_subst_opts (struct subst *cmd);
 static int
 mark_subst_opts(cmd)
   struct subst *cmd;
@@ -607,18 +586,24 @@ mark_subst_opts(cmd)
       {
       case 'i':	/* GNU extension */
       case 'I':	/* GNU extension */
+	if (posixicity == POSIXLY_BASIC)
+	  bad_prog(_(UNKNOWN_S_OPT));
 	flags |= REG_ICASE;
 	break;
 
 #ifdef REG_PERL
       case 's':	/* GNU extension */
       case 'S':	/* GNU extension */
+	if (posixicity == POSIXLY_BASIC)
+	  bad_prog(_(UNKNOWN_S_OPT));
 	if (extended_regexp_flags & REG_PERL)
 	  flags |= REG_DOTALL;
 	break;
 
       case 'x':	/* GNU extension */
       case 'X':	/* GNU extension */
+	if (posixicity == POSIXLY_BASIC)
+	  bad_prog(_(UNKNOWN_S_OPT));
 	if (extended_regexp_flags & REG_PERL)
 	  flags |= REG_EXTENDED;
 	break;
@@ -626,6 +611,8 @@ mark_subst_opts(cmd)
 
       case 'm':	/* GNU extension */
       case 'M':	/* GNU extension */
+	if (posixicity == POSIXLY_BASIC)
+	  bad_prog(_(UNKNOWN_S_OPT));
 	flags |= REG_NEWLINE;
 	break;
 
@@ -646,7 +633,7 @@ mark_subst_opts(cmd)
 	break;
 
       case 'w':
-	cmd->outf = get_openfile(&file_write, "w", true);
+	cmd->outf = get_openfile(&file_write, write_mode, true);
 	return flags;
 
       case '0': case '1': case '2': case '3': case '4':
@@ -680,7 +667,7 @@ mark_subst_opts(cmd)
 
 
 /* read in a label for a `:', `b', or `t' command */
-static char *read_label P_((void));
+static char *read_label (void);
 static char *
 read_label()
 {
@@ -707,7 +694,7 @@ read_label()
    compilation is complete, or a reference created by a `{' to be
    backpatched when the corresponding `}' is found.  */
 static struct sed_label *setup_label
-  P_((struct sed_label *, countT, char *, const struct error_info *));
+  (struct sed_label *, countT, char *, const struct error_info *);
 static struct sed_label *
 setup_label(list, idx, name, err_info)
   struct sed_label *list;
@@ -719,12 +706,12 @@ setup_label(list, idx, name, err_info)
   ret->v_index = idx;
   ret->name = name;
   if (err_info)
-    MEMCPY(&ret->err_info, err_info, sizeof (ret->err_info));
+    memcpy(&ret->err_info, err_info, sizeof (ret->err_info));
   ret->next = list;
   return ret;
 }
 
-static struct sed_label *release_label P_((struct sed_label *list_head));
+static struct sed_label *release_label (struct sed_label *list_head);
 static struct sed_label *
 release_label(list_head)
   struct sed_label *list_head;
@@ -735,22 +722,17 @@ release_label(list_head)
     return NULL;
   ret = list_head->next;
 
-  FREE(list_head->name);
+  free(list_head->name);
 
 #if 0
   /* We use obstacks */
-  FREE(list_head);
+  free(list_head);
 #endif
   return ret;
 }
 
-static struct replacement *new_replacement P_((char *, size_t,
-					       enum replacement_types));
 static struct replacement *
-new_replacement(text, length, type)
-  char *text;
-  size_t length;
-  enum replacement_types type;
+new_replacement(char *text, size_t length, enum replacement_types type)
 {
   struct replacement *r = OB_MALLOC(&obs, 1, struct replacement);
 
@@ -763,7 +745,7 @@ new_replacement(text, length, type)
   return r;
 }
 
-static void setup_replacement P_((struct subst *, const char *, size_t));
+static void setup_replacement (struct subst *, const char *, size_t);
 static void
 setup_replacement(sub, text, length)
      struct subst *sub;
@@ -790,14 +772,23 @@ setup_replacement(sub, text, length)
 	{
 	  /* Preceding the backslash may be some literal text: */
 	  tail = tail->next =
-	    new_replacement(base, CAST(size_t)(p - base), repl_type);
+	    new_replacement(base, (size_t)(p - base), repl_type);
 
 	  repl_type = save_type;
 
 	  /* Skip the backslash and look for a numeric back-reference,
 	     or a case-munging escape if not in POSIX mode: */
 	  ++p;
-	  if (p < text_end && (posixicity != POSIXLY_BASIC || ISDIGIT (*p)))
+	  if (p == text_end)
+	    ++tail->prefix_length;
+
+	  else if (posixicity == POSIXLY_BASIC && !ISDIGIT (*p))
+	    {
+	      p[-1] = *p;
+	      ++tail->prefix_length;
+	    }
+
+	  else
 	    switch (*p)
 	      {
 	      case '0': case '1': case '2': case '3': case '4': 
@@ -843,7 +834,7 @@ setup_replacement(sub, text, length)
 	{
 	  /* Preceding the ampersand may be some literal text: */
 	  tail = tail->next =
-	    new_replacement(base, CAST(size_t)(p - base), repl_type);
+	    new_replacement(base, (size_t)(p - base), repl_type);
 
 	  repl_type = save_type;
 	  tail->subst_id = 0;
@@ -853,13 +844,13 @@ setup_replacement(sub, text, length)
   /* There may be some trailing literal text: */
   if (base < text_end)
     tail = tail->next =
-      new_replacement(base, CAST(size_t)(text_end - base), repl_type);
+      new_replacement(base, (size_t)(text_end - base), repl_type);
 
   tail->next = NULL;
   sub->replacement = root.next;
 }
 
-static void read_text P_((struct text_buf *buf, int leadin_ch));
+static void read_text (struct text_buf *buf, int leadin_ch);
 static void
 read_text(buf, leadin_ch)
   struct text_buf *buf;
@@ -919,7 +910,7 @@ read_text(buf, leadin_ch)
    return non-zero and store the resulting address in `*addr'.
    If the input doesn't look like an address read nothing
    and return zero.  */
-static bool compile_address P_((struct addr *addr, int ch));
+static bool compile_address (struct addr *addr, int ch);
 static bool
 compile_address(addr, ch)
   struct addr *addr;
@@ -943,6 +934,8 @@ compile_address(addr, ch)
       for(;;)
 	{
 	  ch = in_nonblank();
+	  if (posixicity == POSIXLY_BASIC)
+	    goto posix_address_modifier;
           switch(ch)
 	    {
 	    case 'I':	/* GNU extension */
@@ -966,6 +959,7 @@ compile_address(addr, ch)
 	      break;
 
 	    default:
+	    posix_address_modifier:
 	      savchar (ch);
 	      addr->addr_regex = compile_regex (b, flags, 0);
 	      free_buffer(b);
@@ -978,7 +972,7 @@ compile_address(addr, ch)
       addr->addr_number = in_integer(ch);
       addr->addr_type = ADDR_IS_NUM;
       ch = in_nonblank();
-      if (ch != '~')
+      if (ch != '~' || posixicity == POSIXLY_BASIC)
 	{
 	  savchar(ch);
 	}
@@ -992,7 +986,7 @@ compile_address(addr, ch)
 	    }
 	}
     }
-  else if (ch == '+' || ch == '~')
+  else if ((ch == '+' || ch == '~') && posixicity != POSIXLY_BASIC)
     {
       addr->addr_step = in_integer(in_nonblank());
       if (addr->addr_step==0)
@@ -1014,7 +1008,7 @@ compile_address(addr, ch)
 
 /* Read a program (or a subprogram within `{' `}' pairs) in and store
    the compiled form in `*vector'.  Return a pointer to the new vector.  */
-static struct vector *compile_program P_((struct vector *));
+static struct vector *compile_program (struct vector *);
 static struct vector *
 compile_program(vector)
   struct vector *vector;
@@ -1062,9 +1056,10 @@ compile_program(vector)
 	      ch = in_nonblank();
 	    }
 
-	  if (cur_cmd->a1->addr_type == ADDR_IS_NUM
-	      && cur_cmd->a1->addr_number == 0
-	      && (!cur_cmd->a2 || cur_cmd->a2->addr_type != ADDR_IS_REGEX))
+	  if ((cur_cmd->a1->addr_type == ADDR_IS_NUM
+	       && cur_cmd->a1->addr_number == 0)
+	      && ((!cur_cmd->a2 || cur_cmd->a2->addr_type != ADDR_IS_REGEX)
+		  || posixicity == POSIXLY_BASIC))
 	    bad_prog(_(INVALID_LINE_0));
 	}
       if (ch == '!')
@@ -1078,11 +1073,11 @@ compile_program(vector)
       /* Do not accept extended commands in --posix mode.  Also,
 	 a few commands only accept one address in that mode.  */
       if (posixicity == POSIXLY_BASIC)
-	switch (ch)
-	  {
-	    case 'v': case 'L': case 'Q': case 'T':
-	    case 'R': case 'W':
-	      bad_command(ch);
+       switch (ch)
+         {
+           case 'e': case 'F': case 'v': case 'z': case 'L':
+           case 'Q': case 'T': case 'R': case 'W':
+             bad_command(ch);
 
 	    case 'a': case 'i': case 'l':
 	    case '=': case 'r':
@@ -1165,6 +1160,8 @@ compile_program(vector)
 	    ch = inchar();
 	  else
 	    {
+	      if (posixicity == POSIXLY_BASIC)
+		bad_prog(_(EXPECTED_SLASH));
 	      savchar(ch);
 	      ch = '\n';
 	    }
@@ -1193,7 +1190,7 @@ compile_program(vector)
 	case 'L':
 	case 'l':
 	  ch = in_nonblank();
-	  if (ISDIGIT(ch)) 
+	  if (ISDIGIT(ch) && posixicity != POSIXLY_BASIC)
 	    {
 	      cur_cmd->x.int_arg = in_integer(ch);
 	      ch = in_nonblank();
@@ -1208,17 +1205,19 @@ compile_program(vector)
 
 	  break;
 
-	case '=':
-	case 'd':
-	case 'D':
-	case 'g':
-	case 'G':
-	case 'h':
+       case '=':
+       case 'd':
+       case 'D':
+       case 'F':
+       case 'g':
+       case 'G':
+       case 'h':
 	case 'H':
 	case 'n':
 	case 'N':
 	case 'p':
 	case 'P':
+	case 'z':
 	case 'x':
 	  ch = in_nonblank();
 	  if (ch == CLOSE_BRACE || ch == '#')
@@ -1234,12 +1233,12 @@ compile_program(vector)
 	  break;
 
         case 'R':
-	  cur_cmd->x.fp = get_openfile(&file_read, "r", false)->fp;
+	  cur_cmd->x.fp = get_openfile(&file_read, read_mode, false)->fp;
 	  break;
 
         case 'W':
 	case 'w':
-	  cur_cmd->x.outf = get_openfile(&file_write, "w", true);
+	  cur_cmd->x.outf = get_openfile(&file_write, write_mode, true);
 	  break;
 
 	case 's':
@@ -1347,8 +1346,9 @@ compile_program(vector)
               }
             else
               {
-	        char *translate = OB_MALLOC(&obs, YMAP_LENGTH, char);
-                unsigned char *ustring = CAST(unsigned char *)src_buf;
+	        unsigned char *translate =
+		  OB_MALLOC(&obs, YMAP_LENGTH, unsigned char);
+                unsigned char *ustring = (unsigned char *)src_buf;
 
 		if (len != dest_len)
                   bad_prog(_(Y_CMD_LEN));
@@ -1357,7 +1357,7 @@ compile_program(vector)
 	          translate[len] = len;
 
                 while (dest_len--)
-                  translate[(unsigned char)*ustring++] = *dest_buf++;
+                  translate[*ustring++] = (unsigned char)*dest_buf++;
 
 	        cur_cmd->x.translate = translate;
 	      }
@@ -1382,6 +1382,8 @@ compile_program(vector)
       /* this is buried down here so that "continue" statements will miss it */
       ++vector->v_length;
     }
+  if (posixicity == POSIXLY_BASIC && pending_text)
+    bad_prog (_(INCOMPLETE_CMD));
   return vector;
 }
 
@@ -1396,6 +1398,8 @@ normalize_text(buf, len, buftype)
   const char *bufend = buf + len;
   char *p = buf;
   char *q = buf;
+  char ch;
+  int base;
 
   /* This variable prevents normalizing text within bracket
      subexpressions when conforming to POSIX.  If 0, we
@@ -1411,7 +1415,6 @@ normalize_text(buf, len, buftype)
 
   while (p < bufend)
     {
-      int c;
       mbclen = MBRLEN (p, bufend - p, &cur_stat);
       if (mbclen != 1)
 	{
@@ -1427,7 +1430,7 @@ normalize_text(buf, len, buftype)
 	}
 
       if (*p == '\\' && p+1 < bufend && bracket_state == 0)
-	switch ( (c = *++p) )
+	switch (*++p)
 	  {
 #if defined __STDC__ && __STDC__-0
 	  case 'a': *q++ = '\a'; p++; continue;
@@ -1443,14 +1446,12 @@ normalize_text(buf, len, buftype)
 	  case 'v': *q++ = '\v'; p++; continue;
 
 	  case 'd': /* decimal byte */
-	    p = convert_number(q, p+1, bufend, 10, 3, 'd');
-	    q++;
-	    continue;
+            base = 10;
+            goto convert;
 
 	  case 'x': /* hexadecimal byte */
-	    p = convert_number(q, p+1, bufend, 16, 2, 'x');
-	    q++;
-	    continue;
+            base = 16;
+            goto convert;
 
 #ifdef REG_PERL
 	  case '0': case '1': case '2': case '3':
@@ -1459,8 +1460,8 @@ normalize_text(buf, len, buftype)
 		&& p+1 < bufend
 		&& p[1] >= '0' && p[1] <= '9')
 	      {
-		p = convert_number(q, p, bufend, 8, 3, *p);
-		q++;
+                base = 8;
+                goto convert;
 	      }
 	    else
 	      {
@@ -1474,8 +1475,8 @@ normalize_text(buf, len, buftype)
 	  case 'o': /* octal byte */
 	    if (!(extended_regexp_flags & REG_PERL))
 	      {
-	        p = convert_number(q, p+1, bufend,  8, 3, 'o');
-		q++;
+                base = 8;
+                goto convert;
 	      }
 	    else
 	      {
@@ -1487,15 +1488,21 @@ normalize_text(buf, len, buftype)
 	    continue;
 #else
 	  case 'o': /* octal byte */
-	    p = convert_number(q, p+1, bufend,  8, 3, 'o');
-	    q++;
-	    continue;
+            base = 8;
 #endif
+convert:
+            p = convert_number(&ch, p, bufend, base);
+
+            /* for an ampersand in a replacement, pass the \ up one level */
+            if (buftype == TEXT_REPLACEMENT && ch == '&')
+              *q++ = '\\';
+            *q++ = ch;
+            continue;
 
 	  case 'c':
 	    if (++p < bufend)
 	      {
-		*q++ = toupper(*p) ^ 0x40;
+		*q++ = toupper((unsigned char) *p) ^ 0x40;
 		p++;
 		continue;
 	      }
@@ -1556,7 +1563,7 @@ compile_string(cur_program, str, len)
   struct vector *ret;
 
   prog.file = NULL;
-  prog.base = CAST(unsigned char *)str;
+  prog.base = (unsigned char *)str;
   prog.cur = prog.base;
   prog.end = prog.cur + len;
 
@@ -1581,12 +1588,17 @@ compile_file(cur_program, cmdfile)
   struct vector *cur_program;
   const char *cmdfile;
 {
-  size_t len;
   struct vector *ret;
 
   prog.file = stdin;
   if (cmdfile[0] != '-' || cmdfile[1] != '\0')
-    prog.file = ck_fopen(cmdfile, "rt", true);
+    {
+#ifdef HAVE_FOPEN_RT
+      prog.file = ck_fopen(cmdfile, "rt", true);
+#else
+      prog.file = ck_fopen(cmdfile, "r", true);
+#endif
+    }
 
   cur_input.line = 1;
   cur_input.name = cmdfile;
@@ -1615,7 +1627,7 @@ check_final_program(program)
   if (blocks)
     {
       /* update info for error reporting: */
-      MEMCPY(&cur_input, &blocks->err_info, sizeof (cur_input));
+      memcpy(&cur_input, &blocks->err_info, sizeof (cur_input));
       bad_prog(_(EXCESS_OPEN_BRACE));
     }
 
@@ -1623,8 +1635,9 @@ check_final_program(program)
   if (pending_text)
     {
       old_text_buf->text_length = size_buffer(pending_text);
-      old_text_buf->text = MEMDUP(get_buffer(pending_text),
-				  old_text_buf->text_length, char);
+      if (old_text_buf->text_length)
+        old_text_buf->text = MEMDUP(get_buffer(pending_text),
+				    old_text_buf->text_length, char);
       free_buffer(pending_text);
       pending_text = NULL;
     }
@@ -1658,14 +1671,14 @@ check_final_program(program)
     for (p=file_read; p; p=p->link)
       if (p->name)
 	{
-	  FREE(p->name);
+	  free(p->name);
 	  p->name = NULL;
 	}
 
     for (p=file_write; p; p=p->link)
       if (p->name)
 	{
-	  FREE(p->name);
+	  free(p->name);
 	  p->name = NULL;
 	}
   }
@@ -1698,7 +1711,7 @@ finish_program(program)
 	q = p->link;
 #if 0
 	/* We use obstacks. */
-	FREE(p);
+	free(p);
 #endif
       }
 
@@ -1709,7 +1722,7 @@ finish_program(program)
 	q = p->link;
 #if 0
 	/* We use obstacks. */
-	FREE(p);
+	free(p);
 #endif
       }
     file_read = file_write = NULL;
